import sys

from .config import VALUE, FEATURE_FLAGS, BOOL_VALUE, STRING_VALUE, INT_VALUE, DOUBLE_VALUE, SettingType, SETTING_TYPE, \
    UNSUPPORTED_VALUE
from .overridedatasource import OverrideDataSource, FlagOverrides
from .utils import unicode_to_utf8


class LocalDictionaryFlagOverrides(FlagOverrides):
    def __init__(self, source, override_behaviour):
        self.source = source
        self.override_behaviour = override_behaviour

        if sys.version_info[0] == 2:
            self.source = unicode_to_utf8(self.source)  # On Python 2.7, convert unicode to utf-8

    def create_data_source(self, log):
        return LocalDictionaryDataSource(self.source, self.override_behaviour, log)


class LocalDictionaryDataSource(OverrideDataSource):
    def __init__(self, source, override_behaviour, log):
        OverrideDataSource.__init__(self, override_behaviour=override_behaviour)
        self.log = log
        self._config = {}
        for key, value in source.items():
            if isinstance(value, bool):
                value_type = BOOL_VALUE
            elif isinstance(value, str):
                value_type = STRING_VALUE
            elif isinstance(value, int):
                value_type = INT_VALUE
            elif isinstance(value, float):
                value_type = DOUBLE_VALUE
            else:
                value_type = UNSUPPORTED_VALUE

            if FEATURE_FLAGS not in self._config:
                self._config[FEATURE_FLAGS] = {}

            self._config[FEATURE_FLAGS][key] = {VALUE: {value_type: value}}
            setting_type = SettingType.from_type(type(value))
            if setting_type is not None:
                self._config[FEATURE_FLAGS][key][SETTING_TYPE] = int(setting_type)

    def get_overrides(self):
        return self._config



from abc import ABCMeta, abstractmethod


class PollingMode(object):
    """
    Describes a polling mode.
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def identifier(self):
        """
        :return: the identifier of polling mode. Used for analytical purposes in HTTP User-Agent headers.
        """

    @staticmethod
    def auto_poll(poll_interval_seconds=60, max_init_wait_time_seconds=5):
        """
        Creates a configured auto polling configuration.

        :param poll_interval_seconds:
            sets at least how often this policy should fetch the latest configuration and refresh the cache.
        :param max_init_wait_time_seconds:
            sets the maximum waiting time between initialization and the first config acquisition in seconds.
        """

        if poll_interval_seconds < 1:
            poll_interval_seconds = 1

        if max_init_wait_time_seconds < 0:
            max_init_wait_time_seconds = 0

        return AutoPollingMode(poll_interval_seconds=poll_interval_seconds,
                               max_init_wait_time_seconds=max_init_wait_time_seconds)

    @staticmethod
    def lazy_load(cache_refresh_interval_seconds=60):
        """
        Creates a configured lazy loading polling configuration.

        :param cache_refresh_interval_seconds:
            sets how long the cache will store its value before fetching the latest from the network again.
        """

        if cache_refresh_interval_seconds < 1:
            cache_refresh_interval_seconds = 1

        return LazyLoadingMode(cache_refresh_interval_seconds=cache_refresh_interval_seconds)

    @staticmethod
    def manual_poll():
        """
        Creates a configured manual polling configuration.
        """
        return ManualPollingMode()


class AutoPollingMode(PollingMode):
    def __init__(self, poll_interval_seconds, max_init_wait_time_seconds):
        self.poll_interval_seconds = poll_interval_seconds
        self.max_init_wait_time_seconds = max_init_wait_time_seconds

    def identifier(self):
        return "a"


class LazyLoadingMode(PollingMode):
    def __init__(self, cache_refresh_interval_seconds):
        self.cache_refresh_interval_seconds = cache_refresh_interval_seconds

    def identifier(self):
        return "l"


class ManualPollingMode(PollingMode):
    def identifier(self):
        return "m"



import codecs
import sys

from .config import fixup_config_salt_and_segments, VALUE, FEATURE_FLAGS, BOOL_VALUE, STRING_VALUE, \
    INT_VALUE, DOUBLE_VALUE, SettingType, SETTING_TYPE, UNSUPPORTED_VALUE
from .overridedatasource import OverrideDataSource, FlagOverrides
import json
import os

from .utils import unicode_to_utf8


class LocalFileFlagOverrides(FlagOverrides):
    def __init__(self, file_path, override_behaviour):
        self.file_path = file_path
        self.override_behaviour = override_behaviour

    def create_data_source(self, log):
        return LocalFileDataSource(self.file_path, self.override_behaviour, log)


def open_file(file_path, mode='r'):
    # Python 2.7, utf-8 is not supported in open() function
    if sys.version_info[0] == 2:
        return codecs.open(file_path, mode, encoding='utf-8')
    else:
        return open(file_path, mode, encoding='utf-8')


class LocalFileDataSource(OverrideDataSource):
    def __init__(self, file_path, override_behaviour, log):
        OverrideDataSource.__init__(self, override_behaviour=override_behaviour)
        self.log = log
        if not os.path.exists(file_path):
            self.log.error('Cannot find the local config file \'%s\'. '
                           'This is a path that your application provided to the ConfigCat SDK '
                           'by passing it to the constructor of the `LocalFileFlagOverrides` class. '
                           'Read more: https://configcat.com/docs/sdk-reference/python/#json-file',
                           file_path, event_id=1300)

        self._file_path = file_path
        self._config = None
        self._cached_file_stamp = 0

    def get_overrides(self):
        self._reload_file_content()
        return self._config

    def _reload_file_content(self):  # noqa: C901
        try:
            stamp = os.stat(self._file_path).st_mtime
            if stamp != self._cached_file_stamp:
                self._cached_file_stamp = stamp
                with open_file(self._file_path) as file:
                    data = json.load(file)

                    if sys.version_info[0] == 2:
                        data = unicode_to_utf8(data)  # On Python 2.7, convert unicode to utf-8

                    if 'flags' in data:
                        self._config = {FEATURE_FLAGS: {}}
                        source = data['flags']
                        for key, value in source.items():
                            if isinstance(value, bool):
                                value_type = BOOL_VALUE
                            elif isinstance(value, str):
                                value_type = STRING_VALUE
                            elif isinstance(value, int):
                                value_type = INT_VALUE
                            elif isinstance(value, float):
                                value_type = DOUBLE_VALUE
                            else:
                                value_type = UNSUPPORTED_VALUE

                            self._config[FEATURE_FLAGS][key] = {VALUE: {value_type: value}}
                            setting_type = SettingType.from_type(type(value))
                            if setting_type is not None:
                                self._config[FEATURE_FLAGS][key][SETTING_TYPE] = int(setting_type)
                    else:
                        fixup_config_salt_and_segments(data)
                        self._config = data
        except OSError:
            self.log.exception('Failed to read the local config file \'%s\'.', self._file_path, event_id=1302)
        except ValueError:
            self.log.exception('Failed to decode JSON from the local config file \'%s\'.', self._file_path, event_id=2302)



import logging
from threading import Lock

from . import utils
from .configservice import ConfigService
from .config import TARGETING_RULES, VARIATION_ID, PERCENTAGE_OPTIONS, FEATURE_FLAGS, SERVED_VALUE, SETTING_TYPE, \
    is_type_mismatch
from .evaluationdetails import EvaluationDetails
from .evaluationlogbuilder import EvaluationLogBuilder
from .interfaces import ConfigCatClientException
from .logger import Logger
from .configfetcher import ConfigFetcher
from .configcache import NullConfigCache
from .configcatoptions import ConfigCatOptions, Hooks
from .overridedatasource import OverrideBehaviour
from .refreshresult import RefreshResult
from .rolloutevaluator import RolloutEvaluator, get_value
from collections import namedtuple
import copy
from .utils import method_is_called_from, get_date_time
import re

KeyValue = namedtuple('KeyValue', 'key value')


class ConfigCatClient(object):
    _lock = Lock()
    _instances = {}

    @classmethod
    def get(cls, sdk_key, options=None):
        """
        Creates a new or gets an already existing `ConfigCatClient` for the given `sdk_key`.

        :param sdk_key: ConfigCat SDK Key to access your configuration.
        :param options: Configuration `ConfigCatOptions` for `ConfigCatClient`.
        :return: the `ConfigCatClient` instance.
        """
        with cls._lock:
            client = cls._instances.get(sdk_key)
            if client is not None:
                if options is not None:
                    client.log.warning('There is an existing client instance for the specified SDK Key. '
                                       'No new client instance will be created and the specified options are ignored. '
                                       'Returning the existing client instance. SDK Key: \'%s\'.',
                                       sdk_key, event_id=3000)
                return client

            if options is None:
                options = ConfigCatOptions()

            client = ConfigCatClient(sdk_key=sdk_key,
                                     options=options)
            cls._instances[sdk_key] = client
            return client

    @classmethod
    def close_all(cls):
        """
        Closes all ConfigCatClient instances.
        """
        with cls._lock:
            for key, value in list(cls._instances.items()):
                value._close_resources()
            cls._instances.clear()

    def __init__(self,
                 sdk_key,
                 options=ConfigCatOptions()):

        self._hooks = options.hooks if options.hooks is not None else Hooks()
        self.log = Logger('configcat', self._hooks)

        if not method_is_called_from(ConfigCatClient.get):
            raise ConfigCatClientException('ConfigCatClient.__init__() is private. Create the ConfigCat Client as a '
                                           'Singleton object with `ConfigCatClient.get()` instead.')

        if sdk_key is None:
            raise ConfigCatClientException('SDK Key is required.')

        if options.flag_overrides:
            self._override_data_source = options.flag_overrides.create_data_source(self.log)
        else:
            self._override_data_source = None

        # In case of local only flag overrides mode, we accept any SDK Key format.
        if self._override_data_source is None or self._override_data_source.get_behaviour() != OverrideBehaviour.LocalOnly:
            is_valid_sdk_key = re.match('^.{22}/.{22}$', sdk_key) is not None or \
                re.match('^configcat-sdk-1/.{22}/.{22}$', sdk_key) is not None or \
                (options.base_url and re.match('^configcat-proxy/.+$', sdk_key) is not None)
            if not is_valid_sdk_key:
                raise ConfigCatClientException('SDK Key `%s` is invalid.' % sdk_key)

        self._sdk_key = sdk_key
        self._default_user = options.default_user
        self._rollout_evaluator = RolloutEvaluator(self.log)
        config_cache = options.config_cache if options.config_cache is not None else NullConfigCache()

        if self._override_data_source and self._override_data_source.get_behaviour() == OverrideBehaviour.LocalOnly:
            self._config_fetcher = None
            self._config_service = None
        else:
            self._config_fetcher = ConfigFetcher(self._sdk_key,
                                                 self.log,
                                                 options.polling_mode.identifier(),
                                                 options.base_url,
                                                 options.proxies, options.proxy_auth,
                                                 options.connect_timeout_seconds, options.read_timeout_seconds,
                                                 options.data_governance)
            self._config_service = ConfigService(self._sdk_key,
                                                 options.polling_mode,
                                                 self._hooks,
                                                 self._config_fetcher,
                                                 self.log,
                                                 config_cache,
                                                 options.offline)

    def get_value(self, key, default_value, user=None):
        """
        Gets the value of a feature flag or setting identified by the given `key`.

        :param key: the identifier of the feature flag or setting.
        :param default_value: in case of any failure, this value will be returned.
        :param user: the user object to identify the caller.
        :return: the value.
        """
        config, fetch_time = self._get_config()
        if config is None or config.get(FEATURE_FLAGS) is None:
            message = 'Config JSON is not present when evaluating setting \'%s\'. ' \
                      'Returning the `%s` parameter that you specified in your application: \'%s\'.'
            message_args = (key, 'default_value', str(default_value))
            self.log.error(message, *message_args, event_id=1000)
            self._hooks.invoke_on_flag_evaluated(
                EvaluationDetails.from_error(key, default_value, Logger.format(message, message_args)))
            return default_value

        details = self._evaluate(key=key,
                                 user=user,
                                 default_value=default_value,
                                 default_variation_id=None,
                                 config=config,
                                 fetch_time=fetch_time)

        return details.value

    def get_value_details(self, key, default_value, user=None):
        """
        Gets the value and evaluation details of a feature flag or setting identified by the given `key`.

        :param key: the identifier of the feature flag or setting.
        :param default_value: in case of any failure, this value will be returned.
        :param user: the user object to identify the caller.
        :return: the evaluation details.
        """
        config, fetch_time = self._get_config()
        if config is None or config.get(FEATURE_FLAGS) is None:
            message = 'Config JSON is not present when evaluating setting \'%s\'. ' \
                      'Returning the `%s` parameter that you specified in your application: \'%s\'.'
            message_args = (key, 'default_value', str(default_value))
            self.log.error(message, *message_args, event_id=1000)
            details = EvaluationDetails.from_error(key, default_value, Logger.format(message, message_args))
            self._hooks.invoke_on_flag_evaluated(details)
            return details

        details = self._evaluate(key=key,
                                 user=user,
                                 default_value=default_value,
                                 default_variation_id=None,
                                 config=config,
                                 fetch_time=fetch_time)

        return details

    def get_all_keys(self):
        """
        Gets all setting keys.

        :return: list of keys.
        """
        config, _ = self._get_config()
        if config is None:
            self.log.error('Config JSON is not present. Returning %s.', 'empty list', event_id=1000)
            return []

        settings = config.get(FEATURE_FLAGS, {})
        return list(settings)

    def get_key_and_value(self, variation_id):  # noqa: C901
        """
        Gets the key of a setting, and it's value identified by the given Variation ID (analytics)

        :param variation_id: variation ID
        :return: key and value
        """
        config, _ = self._get_config()
        if config is None:
            self.log.error('Config JSON is not present. Returning %s.', 'None', event_id=1000)
            return None

        settings = config.get(FEATURE_FLAGS, {})
        try:
            for key, value in list(settings.items()):
                setting_type = value.get(SETTING_TYPE)
                if variation_id == value.get(VARIATION_ID):
                    return KeyValue(key, get_value(value, setting_type))

                targeting_rules = value.get(TARGETING_RULES, [])
                for targeting_rule in targeting_rules:
                    served_value = targeting_rule.get(SERVED_VALUE)
                    if served_value is not None:
                        if variation_id == served_value.get(VARIATION_ID):
                            return KeyValue(key, get_value(served_value, setting_type))
                    else:
                        targeting_rule_percentage_options = targeting_rule.get(PERCENTAGE_OPTIONS, [])
                        for percentage_option in targeting_rule_percentage_options:
                            if variation_id == percentage_option.get(VARIATION_ID):
                                return KeyValue(key, get_value(percentage_option, setting_type))

                percentage_options = value.get(PERCENTAGE_OPTIONS, [])
                for percentage_option in percentage_options:
                    if variation_id == percentage_option.get(VARIATION_ID):
                        return KeyValue(key, get_value(percentage_option, setting_type))
        except Exception:
            self.log.exception('Error occurred in the `' + __name__ + '` method. Returning None.', event_id=1002)
            return None

        self.log.error('Could not find the setting for the specified variation ID: \'%s\'.', variation_id, event_id=2011)
        return None

    def get_all_values(self, user=None):
        """
        Evaluates and returns the values of all feature flags and settings.

        :param user: the user object to identify the caller.
        :return: dictionary of values
        """
        config, _ = self._get_config()
        if config is None:
            self.log.error('Config JSON is not present. Returning %s.', 'empty dictionary', event_id=1000)
            return {}

        settings = config.get(FEATURE_FLAGS, {})
        all_values = {}
        for key in list(settings):
            value = self.get_value(key, None, user)
            if value is not None:
                all_values[key] = value

        return all_values

    def get_all_value_details(self, user=None):
        """
        Gets the values along with evaluation details of all feature flags and settings.

        :param user: the user object to identify the caller.
        :return: list of all evaluation details
        """
        config, fetch_time = self._get_config()
        if config is None:
            self.log.error('Config JSON is not present. Returning %s.', 'empty list', event_id=1000)
            return []

        details_result = []
        settings = config.get(FEATURE_FLAGS, {})
        for key in list(settings):
            details = self._evaluate(key=key,
                                     user=user,
                                     default_value=None,
                                     default_variation_id=None,
                                     config=config,
                                     fetch_time=fetch_time)
            details_result.append(details)

        return details_result

    def force_refresh(self):
        """
        Initiates a force refresh on the cached configuration.

        :return: RefreshResult object
        """
        if self._config_service:
            return self._config_service.refresh()

        return RefreshResult(False,
                             'The SDK uses the LocalOnly flag override behavior which prevents making HTTP requests.')

    def set_default_user(self, user):
        """
        Sets the default user.

        :param user: the user object to identify the caller.
        """
        self._default_user = user

    def clear_default_user(self):
        """
        Sets the default user to None.
        """
        self._default_user = None

    def set_online(self):
        """
        Configures the SDK to allow HTTP requests.
        """
        if self._config_service:
            self._config_service.set_online()
        else:
            self.log.warning('Client is configured to use the `%s` override behavior, thus `%s()` has no effect.',
                             'LocalOnly', 'set_online', event_id=3202)

    def set_offline(self):
        """
        Configures the SDK to not initiate HTTP requests and work only from its cache.
        """
        if self._config_service:
            self._config_service.set_offline()

    def is_offline(self):
        """
        True when the SDK is configured not to initiate HTTP requests, otherwise False.
        """
        if self._config_service:
            return self._config_service.is_offline()

        return True

    def get_hooks(self):
        """
        Gets the Hooks object for subscribing events.

        :return: Hooks object
        """
        return self._hooks

    def close(self):
        """
        Closes the underlying resources.
        """
        with ConfigCatClient._lock:
            self._close_resources()
            ConfigCatClient._instances.pop(self._sdk_key)

    def _close_resources(self):
        if self._config_service:
            self._config_service.close()
        self._hooks.clear()

    def _get_config(self):
        if self._override_data_source:
            behaviour = self._override_data_source.get_behaviour()

            if behaviour == OverrideBehaviour.LocalOnly:
                return self._override_data_source.get_overrides(), utils.distant_past
            elif behaviour == OverrideBehaviour.RemoteOverLocal:
                remote_config, fetch_time = self._config_service.get_config()
                local_config = self._override_data_source.get_overrides()
                if not remote_config:
                    remote_config = {FEATURE_FLAGS: {}}
                if not local_config:
                    local_config = {FEATURE_FLAGS: {}}
                result = copy.deepcopy(local_config)
                result[FEATURE_FLAGS].update(remote_config[FEATURE_FLAGS])
                return result, fetch_time
            elif behaviour == OverrideBehaviour.LocalOverRemote:
                remote_config, fetch_time = self._config_service.get_config()
                local_config = self._override_data_source.get_overrides()
                if not remote_config:
                    remote_config = {FEATURE_FLAGS: {}}
                if not local_config:
                    local_config = {FEATURE_FLAGS: {}}
                result = copy.deepcopy(remote_config)
                result[FEATURE_FLAGS].update(local_config[FEATURE_FLAGS])
                return result, fetch_time

        return self._config_service.get_config()

    def _check_type_mismatch(self, value, default_value):
        if default_value is not None and is_type_mismatch(value, type(default_value)):
            self.log.warning("The type of a setting does not match the type of the specified default value (%s). "
                             "Setting's type was %s but the default value's type was %s. "
                             "Please make sure that using a default value not matching the setting's type was intended." %
                             (default_value, type(value), type(default_value)), event_id=4002)

    def _evaluate(self, key, user, default_value, default_variation_id, config, fetch_time):
        user = user if user is not None else self._default_user

        # Skip building the evaluation log if it won't be logged.
        log_builder = EvaluationLogBuilder() if self.log.isEnabledFor(logging.INFO) else None

        value, variation_id, rule, percentage_rule, error = self._rollout_evaluator.evaluate(
            key=key,
            user=user,
            default_value=default_value,
            default_variation_id=default_variation_id,
            config=config,
            log_builder=log_builder)

        self._check_type_mismatch(value, default_value)

        if log_builder:
            self.log.info(str(log_builder), event_id=5000)

        details = EvaluationDetails(key=key,
                                    value=value,
                                    variation_id=variation_id,
                                    fetch_time=get_date_time(fetch_time) if fetch_time is not None else None,
                                    user=user,
                                    is_default_value=True if error else False,
                                    error=error,
                                    matched_targeting_rule=rule,
                                    matched_percentage_option=percentage_rule)
        self._hooks.invoke_on_flag_evaluated(details)
        return details



import json

__PREDEFINED__ = ['Identifier', 'Email', 'Country']

from collections import OrderedDict
from datetime import datetime


class User(object):
    """
    User Object. Contains user attributes which are used for evaluating targeting rules and percentage options.
    """

    def __init__(self, identifier, email=None, country=None, custom=None):
        """
        Initialize a User object.

        Args:
            identifier: The unique identifier of the user or session (e.g. email address, primary key, session ID, etc.)
            email: Email address of the user.
            country: Country of the user.
            custom: Custom attributes of the user for advanced targeting rule definitions (e.g. role, subscription type, etc.)

            All comparators support string values as User Object attribute (in some cases they need to be provided in a
            specific format though, see below), but some of them also support other types of values. It depends on the
            comparator how the values will be handled. The following rules apply:

            Text-based comparators (EQUALS, IS_ONE_OF, etc.)
            * accept string values,
            * all other values are automatically converted to string
              (a warning will be logged but evaluation will continue as normal).

            SemVer-based comparators (IS_ONE_OF_SEMVER, LESS_THAN_SEMVER, GREATER_THAN_SEMVER, etc.)
            * accept string values containing a properly formatted, valid semver value,
            * all other values are considered invalid
              (a warning will be logged and the currently evaluated targeting rule will be skipped).

            Number-based comparators (EQUALS_NUMBER, LESS_THAN_NUMBER, GREATER_THAN_OR_EQUAL_NUMBER, etc.)
            * accept float values and all other numeric values which can safely be converted to float,
            * accept string values containing a properly formatted, valid float value,
            * all other values are considered invalid
              (a warning will be logged and the currently evaluated targeting rule will be skipped).

            Date time-based comparators (BEFORE_DATETIME / AFTER_DATETIME)
            * accept datetime values, which are automatically converted to a second-based Unix timestamp
              (datetime values with naive timezone are considered to be in UTC),
            * accept float values representing a second-based Unix timestamp
              and all other numeric values which can safely be converted to float,
            * accept string values containing a properly formatted, valid float value,
            * all other values are considered invalid
              (a warning will be logged and the currently evaluated targeting rule will be skipped).

            String array-based comparators (ARRAY_CONTAINS_ANY_OF / ARRAY_NOT_CONTAINS_ANY_OF)
            * accept arrays of strings,
            * accept string values containing a valid JSON string which can be deserialized to an array of strings,
            * all other values are considered invalid
              (a warning will be logged and the currently evaluated targeting rule will be skipped).
        """
        self.__identifier = identifier if identifier is not None else ''
        self.__data = {'Identifier': identifier, 'Email': email, 'Country': country}
        self.__custom = custom

    def get_identifier(self):
        return self.__identifier

    def get_attribute(self, attribute):
        attribute = str(attribute)
        if attribute in __PREDEFINED__:
            return self.__data[attribute]

        return self.__custom.get(attribute) if self.__custom else None

    def __str__(self):
        def serializer(obj):
            if isinstance(obj, datetime):
                return obj.isoformat()

            raise TypeError("Type not serializable")

        dump = OrderedDict([
            ('Identifier', self.__identifier),
            ('Email', self.__data.get('Email')),
            ('Country', self.__data.get('Country'))
        ])
        if self.__custom:
            dump.update(self.__custom)

        filtered_dump = OrderedDict([(k, v) for k, v in dump.items() if v is not None])
        return json.dumps(filtered_dump, ensure_ascii=False, separators=(',', ':'), default=serializer)



import logging
import sys


class Logger(logging.LoggerAdapter):
    def __init__(self, name, hooks):
        super(Logger, self).__init__(logging.getLogger(name), {})
        self._hooks = hooks

    def process(self, msg, kwargs):
        # Remove event_id from kwargs (as it's not a built-in argument expected by the logging framework)
        # and put it in the extra dict so users can access it without parsing.
        event_id = kwargs.pop('event_id', 0)
        extra = kwargs.setdefault('extra', {})
        extra['event_id'] = event_id

        # Include the event_id in the message.
        return "[" + str(event_id) + "] " + msg, kwargs

    def error(self, msg, *args, **kwargs):
        self._hooks.invoke_on_error(Logger.format(msg, args))
        super(Logger, self).error(msg, *args, **kwargs)

    def exception(self, msg, *args, **kwargs):
        self._hooks.invoke_on_error(Logger.format(msg, args, sys.exc_info()[1]))
        super(Logger, self).exception(msg, *args, **kwargs)

    @staticmethod
    def format(msg, args, exc=None):
        msg = msg % args if len(args) > 0 else msg
        return msg if exc is None else msg + '\n' + str(exc)



import logging
from .datagovernance import DataGovernance
from .pollingmode import PollingMode


class Hooks(object):
    """
    Events fired by [ConfigCatClient].
    """

    def __init__(self, on_client_ready=None, on_config_changed=None,
                 on_flag_evaluated=None, on_error=None):
        self._on_client_ready_callbacks = [on_client_ready] if on_client_ready else []
        self._on_config_changed_callbacks = [on_config_changed] if on_config_changed else []
        self._on_flag_evaluated_callbacks = [on_flag_evaluated] if on_flag_evaluated else []
        self._on_error_callbacks = [on_error] if on_error else []

    def add_on_client_ready(self, callback):
        self._on_client_ready_callbacks.append(callback)

    def add_on_config_changed(self, callback):
        self._on_config_changed_callbacks.append(callback)

    def add_on_flag_evaluated(self, callback):
        self._on_flag_evaluated_callbacks.append(callback)

    def add_on_error(self, callback):
        self._on_error_callbacks.append(callback)

    def invoke_on_client_ready(self):
        for callback in self._on_client_ready_callbacks:
            try:
                callback()
            except Exception as e:
                error = 'Exception occurred during invoke_on_client_ready callback: ' + str(e)
                self.invoke_on_error(error)
                logging.error(error)

    def invoke_on_config_changed(self, config):
        for callback in self._on_config_changed_callbacks:
            try:
                callback(config)
            except Exception as e:
                error = 'Exception occurred during invoke_on_config_changed callback: ' + str(e)
                self.invoke_on_error(error)
                logging.error(error)

    def invoke_on_flag_evaluated(self, evaluation_details):
        for callback in self._on_flag_evaluated_callbacks:
            try:
                callback(evaluation_details)
            except Exception as e:
                error = 'Exception occurred during invoke_on_flag_evaluated callback: ' + str(e)
                self.invoke_on_error(error)
                logging.error(error)

    def invoke_on_error(self, error):
        for callback in self._on_error_callbacks:
            try:
                callback(error)
            except Exception as e:
                logging.error('Exception occurred during invoke_on_error callback: ' + str(e))

    def clear(self):
        self._on_client_ready_callbacks[:] = []
        self._on_config_changed_callbacks[:] = []
        self._on_flag_evaluated_callbacks[:] = []
        self._on_error_callbacks[:] = []


class ConfigCatOptions(object):
    """
    Configuration options for ConfigCatClient.
    """

    def __init__(self,
                 base_url=None,
                 polling_mode=PollingMode.auto_poll(),
                 config_cache=None,
                 proxies=None,
                 proxy_auth=None,
                 connect_timeout_seconds=10,
                 read_timeout_seconds=30,
                 flag_overrides=None,
                 data_governance=DataGovernance.Global,
                 default_user=None,
                 hooks=None,
                 offline=False):
        # The base ConfigCat CDN url.
        self.base_url = base_url

        # The polling mode.
        self.polling_mode = polling_mode

        # The cache implementation used to cache the downloaded config.json.
        self.config_cache = config_cache

        # Proxy addresses. e.g. { "https": "your_proxy_ip:your_proxy_port" }
        self.proxies = proxies

        # Proxy authentication. e.g. HTTPProxyAuth('username', 'password')
        self.proxy_auth = proxy_auth

        # The number of seconds to wait for the server to make the initial connection
        self.connect_timeout_seconds = connect_timeout_seconds

        # The number of seconds to wait for the server to respond before giving up.
        self.read_timeout_seconds = read_timeout_seconds

        # Feature flag and setting overrides.
        self.flag_overrides = flag_overrides

        # Default: `DataGovernance.Global`. Set this parameter to be in sync with the
        # Data Governance preference on the [Dashboard](https://app.configcat.com/organization/data-governance).
        # (Only Organization Admins have access)
        self.data_governance = data_governance

        # The default user, used as fallback when there's no user parameter is passed to the getValue() method.
        self.default_user = default_user

        # Hooks for events sent by ConfigCatClient.
        self.hooks = hooks

        # Indicates whether the SDK should be initialized in offline mode or not.
        self.offline = offline



from .config import Comparator
from .utils import get_date_time


class EvaluationLogBuilder(object):
    def __init__(self):
        self.indent_level = 0
        self.text = ''

    @staticmethod
    def trunc_comparison_value_if_needed(comparator, comparison_value):
        if comparator in [Comparator.IS_ONE_OF_HASHED,
                          Comparator.IS_NOT_ONE_OF_HASHED,
                          Comparator.EQUALS_HASHED,
                          Comparator.NOT_EQUALS_HASHED,
                          Comparator.STARTS_WITH_ANY_OF_HASHED,
                          Comparator.NOT_STARTS_WITH_ANY_OF_HASHED,
                          Comparator.ENDS_WITH_ANY_OF_HASHED,
                          Comparator.NOT_ENDS_WITH_ANY_OF_HASHED,
                          Comparator.ARRAY_CONTAINS_ANY_OF_HASHED,
                          Comparator.ARRAY_NOT_CONTAINS_ANY_OF_HASHED]:
            if isinstance(comparison_value, list):
                length = len(comparison_value)
                if length > 1:
                    return '[<{} hashed values>]'.format(length)
                return '[<{} hashed value>]'.format(length)

            return "'<hashed value>'"

        if isinstance(comparison_value, list):
            length_limit = 10
            length = len(comparison_value)
            if length > length_limit:
                remaining = length - length_limit
                if remaining == 1:
                    more_text = "<1 more value>"
                else:
                    more_text = "<{} more values>".format(remaining)

                return str(comparison_value[:length_limit])[:-1] + ', ... ' + more_text + ']'

            return str(comparison_value)

        if comparator in [Comparator.BEFORE_DATETIME, Comparator.AFTER_DATETIME]:
            time = get_date_time(comparison_value)
            return "'%s' (%sZ UTC)" % (str(comparison_value), time.strftime('%Y-%m-%dT%H:%M:%S.%f')[:-3])

        return "'%s'" % str(comparison_value)

    def increase_indent(self):
        self.indent_level += 1
        return self

    def decrease_indent(self):
        self.indent_level = max(0, self.indent_level - 1)
        return self

    def append(self, text):
        self.text += text
        return self

    def new_line(self, text=None):
        self.text += '\n' + '  ' * self.indent_level
        if text:
            self.text += text
        return self

    def __str__(self):
        return self.text



import sys

from enum import IntEnum

CONFIG_FILE_NAME = 'config_v6'
SERIALIZATION_FORMAT_VERSION = 'v2'

# Config
PREFERENCES = 'p'
SEGMENTS = 's'
FEATURE_FLAGS = 'f'

# Preferences
BASE_URL = 'u'
REDIRECT = 'r'
SALT = 's'

# Segment
SEGMENT_NAME = 'n'  # The first 4 characters of the Segment's name
SEGMENT_CONDITIONS = 'r'  # The list of segment rule conditions (has a logical AND relation between the items).

# Segment Condition (User Condition)
COMPARISON_ATTRIBUTE = 'a'  # The attribute of the user object that should be used to evaluate this rule
COMPARATOR = 'c'

# Feature flag (Evaluation Formula)
SETTING_TYPE = 't'  # 0 = bool, 1 = string, 2 = int, 3 = double
PERCENTAGE_RULE_ATTRIBUTE = 'a'  # Percentage rule evaluation hashes this attribute of the User object to calculate the buckets
TARGETING_RULES = 'r'  # Targeting Rules (Logically connected by OR)
PERCENTAGE_OPTIONS = 'p'  # Percentage Options without conditions
VALUE = 'v'
VARIATION_ID = 'i'
INLINE_SALT = 'inline_salt'

# Targeting Rule (Evaluation Rule)
CONDITIONS = 'c'
SERVED_VALUE = 's'  # Value and Variation ID
TARGETING_RULE_PERCENTAGE_OPTIONS = 'p'

# Condition
USER_CONDITION = 'u'
SEGMENT_CONDITION = 's'  # Segment targeting rule
PREREQUISITE_FLAG_CONDITION = 'p'  # Prerequisite flag targeting rule

# Segment Condition
SEGMENT_INDEX = 's'
SEGMENT_COMPARATOR = 'c'
INLINE_SEGMENT = 'inline_segment'

# Prerequisite Flag Condition
PREREQUISITE_FLAG_KEY = 'f'
PREREQUISITE_COMPARATOR = 'c'

# Percentage Option
PERCENTAGE = 'p'

# Value
BOOL_VALUE = 'b'
STRING_VALUE = 's'
INT_VALUE = 'i'
DOUBLE_VALUE = 'd'
STRING_LIST_VALUE = 'l'
UNSUPPORTED_VALUE = 'unsupported_value'


def is_type_mismatch(value, py_type):
    is_float_int_mismatch = \
        (type(value) is float and py_type is int) or \
        (type(value) is int and py_type is float)

    # On Python 2.7, ignore the type mismatch between str and unicode.
    # (ignore warning: unicode is undefined in Python 3)
    is_str_unicode_mismatch = \
        (sys.version_info[0] == 2 and type(value) is unicode and py_type is str) or \
        (sys.version_info[0] == 2 and type(value) is str and py_type is unicode)  # noqa: F821

    if type(value) is not py_type:
        if not is_float_int_mismatch and not is_str_unicode_mismatch:
            return True

    return False


def get_value(dictionary, setting_type):
    value_descriptor = dictionary.get(VALUE)
    if value_descriptor is None:
        raise ValueError('Value is missing')

    if setting_type not in list(map(int, SettingType)):
        raise ValueError('Unsupported setting type')

    expected_value_type, expected_py_type = SettingType.get_type_info(setting_type)
    if expected_value_type is None:
        raise ValueError('Unsupported setting type')

    value = value_descriptor.get(expected_value_type)
    if value is None or is_type_mismatch(value, expected_py_type):
        raise ValueError("Setting value is not of the expected type %s" % expected_py_type)

    return value


def get_value_type(dictionary):
    value = dictionary.get(VALUE)
    if value is not None:
        if value.get(BOOL_VALUE) is not None:
            return bool
        if value.get(STRING_VALUE) is not None:
            return str
        if value.get(INT_VALUE) is not None:
            return int
        if value.get(DOUBLE_VALUE) is not None:
            return float

    return None


class SettingType(IntEnum):
    BOOL = 0
    STRING = 1
    INT = 2
    DOUBLE = 3

    @staticmethod
    def get_type_info(setting_type):
        return setting_type_mapping.get(setting_type, (None, None))

    @staticmethod
    def from_type(object_type):
        if object_type is bool:
            return SettingType.BOOL
        if object_type is str:
            return SettingType.STRING
        if object_type is int:
            return SettingType.INT
        if object_type is float:
            return SettingType.DOUBLE

        return None

    @staticmethod
    def to_type(setting_type):
        return SettingType.get_type_info(setting_type)[1]

    @staticmethod
    def to_value_type(setting_type):
        return SettingType.get_type_info(setting_type)[0]


setting_type_mapping = {
    SettingType.BOOL: (BOOL_VALUE, bool),
    SettingType.STRING: (STRING_VALUE, str),
    SettingType.INT: (INT_VALUE, int),
    SettingType.DOUBLE: (DOUBLE_VALUE, float)
}


class PrerequisiteComparator(IntEnum):
    EQUALS = 0
    NOT_EQUALS = 1


class SegmentComparator(IntEnum):
    IS_IN = 0
    IS_NOT_IN = 1


class Comparator(IntEnum):
    IS_ONE_OF = 0
    IS_NOT_ONE_OF = 1
    CONTAINS_ANY_OF = 2
    NOT_CONTAINS_ANY_OF = 3
    IS_ONE_OF_SEMVER = 4
    IS_NOT_ONE_OF_SEMVER = 5
    LESS_THAN_SEMVER = 6
    LESS_THAN_OR_EQUAL_SEMVER = 7
    GREATER_THAN_SEMVER = 8
    GREATER_THAN_OR_EQUAL_SEMVER = 9
    EQUALS_NUMBER = 10
    NOT_EQUALS_NUMBER = 11
    LESS_THAN_NUMBER = 12
    LESS_THAN_OR_EQUAL_NUMBER = 13
    GREATER_THAN_NUMBER = 14
    GREATER_THAN_OR_EQUAL_NUMBER = 15
    IS_ONE_OF_HASHED = 16
    IS_NOT_ONE_OF_HASHED = 17
    BEFORE_DATETIME = 18
    AFTER_DATETIME = 19
    EQUALS_HASHED = 20
    NOT_EQUALS_HASHED = 21
    STARTS_WITH_ANY_OF_HASHED = 22
    NOT_STARTS_WITH_ANY_OF_HASHED = 23
    ENDS_WITH_ANY_OF_HASHED = 24
    NOT_ENDS_WITH_ANY_OF_HASHED = 25
    ARRAY_CONTAINS_ANY_OF_HASHED = 26
    ARRAY_NOT_CONTAINS_ANY_OF_HASHED = 27
    EQUALS = 28
    NOT_EQUALS = 29
    STARTS_WITH_ANY_OF = 30
    NOT_STARTS_WITH_ANY_OF = 31
    ENDS_WITH_ANY_OF = 32
    NOT_ENDS_WITH_ANY_OF = 33
    ARRAY_CONTAINS_ANY_OF = 34
    ARRAY_NOT_CONTAINS_ANY_OF = 35


COMPARATOR_TEXTS = [
    'IS ONE OF',                  # IS_ONE_OF
    'IS NOT ONE OF',              # IS_NOT_ONE_OF
    'CONTAINS ANY OF',            # CONTAINS_ANY_OF
    'NOT CONTAINS ANY OF',        # NOT_CONTAINS_ANY_OF
    'IS ONE OF',                  # IS_ONE_OF_SEMVER
    'IS NOT ONE OF',              # IS_NOT_ONE_OF_SEMVER
    '<',                          # LESS_THAN_SEMVER
    '<=',                         # LESS_THAN_OR_EQUAL_SEMVER
    '>',                          # GREATER_THAN_SEMVER
    '>=',                         # GREATER_THAN_OR_EQUAL_SEMVER
    '=',                          # EQUALS_NUMBER
    '!=',                         # NOT_EQUALS_NUMBER
    '<',                          # LESS_THAN_NUMBER
    '<=',                         # LESS_THAN_OR_EQUAL_NUMBER
    '>',                          # GREATER_THAN_NUMBER
    '>=',                         # GREATER_THAN_OR_EQUAL_NUMBER
    'IS ONE OF',                  # IS_ONE_OF_HASHED
    'IS NOT ONE OF',              # IS_NOT_ONE_OF_HASHED
    'BEFORE',                     # BEFORE_DATETIME
    'AFTER',                      # AFTER_DATETIME
    'EQUALS',                     # EQUALS_HASHED
    'NOT EQUALS',                 # NOT_EQUALS_HASHED
    'STARTS WITH ANY OF',         # STARTS_WITH_ANY_OF_HASHED
    'NOT STARTS WITH ANY OF',     # NOT_STARTS_WITH_ANY_OF_HASHED
    'ENDS WITH ANY OF',           # ENDS_WITH_ANY_OF_HASHED
    'NOT ENDS WITH ANY OF',       # NOT_ENDS_WITH_ANY_OF_HASHED
    'ARRAY CONTAINS ANY OF',      # ARRAY_CONTAINS_ANY_OF_HASHED
    'ARRAY NOT CONTAINS ANY OF',  # ARRAY_NOT_CONTAINS_ANY_OF_HASHED
    'EQUALS',                     # EQUALS
    'NOT EQUALS',                 # NOT_EQUALS
    'STARTS WITH ANY OF',         # STARTS_WITH_ANY_OF
    'NOT STARTS WITH ANY OF',     # NOT_STARTS_WITH_ANY_OF
    'ENDS WITH ANY OF',           # ENDS_WITH_ANY_OF
    'NOT ENDS WITH ANY OF',       # NOT_ENDS_WITH_ANY_OF
    'ARRAY CONTAINS ANY OF',      # ARRAY_CONTAINS_ANY_OF
    'ARRAY NOT CONTAINS ANY OF'   # ARRAY_NOT_CONTAINS_ANY_OF
]
COMPARISON_VALUES = [
    STRING_LIST_VALUE,  # IS_ONE_OF
    STRING_LIST_VALUE,  # IS_NOT_ONE_OF
    STRING_LIST_VALUE,  # CONTAINS_ANY_OF
    STRING_LIST_VALUE,  # NOT_CONTAINS_ANY_OF
    STRING_LIST_VALUE,  # IS_ONE_OF_SEMVER
    STRING_LIST_VALUE,  # IS_NOT_ONE_OF_SEMVER
    STRING_VALUE,       # LESS_THAN_SEMVER
    STRING_VALUE,       # LESS_THAN_OR_EQUAL_SEMVER
    STRING_VALUE,       # GREATER_THAN_SEMVER
    STRING_VALUE,       # GREATER_THAN_OR_EQUAL_SEMVER
    DOUBLE_VALUE,       # EQUALS_NUMBER
    DOUBLE_VALUE,       # NOT_EQUALS_NUMBER
    DOUBLE_VALUE,       # LESS_THAN_NUMBER
    DOUBLE_VALUE,       # LESS_THAN_OR_EQUAL_NUMBER
    DOUBLE_VALUE,       # GREATER_THAN_NUMBER
    DOUBLE_VALUE,       # GREATER_THAN_OR_EQUAL_NUMBER
    STRING_LIST_VALUE,  # IS_ONE_OF_HASHED
    STRING_LIST_VALUE,  # IS_NOT_ONE_OF_HASHED
    DOUBLE_VALUE,       # BEFORE_DATETIME
    DOUBLE_VALUE,       # AFTER_DATETIME
    STRING_VALUE,       # EQUALS_HASHED
    STRING_VALUE,       # NOT_EQUALS_HASHED
    STRING_LIST_VALUE,  # STARTS_WITH_ANY_OF_HASHED
    STRING_LIST_VALUE,  # NOT_STARTS_WITH_ANY_OF_HASHED
    STRING_LIST_VALUE,  # ENDS_WITH_ANY_OF_HASHED
    STRING_LIST_VALUE,  # NOT_ENDS_WITH_ANY_OF_HASHED
    STRING_LIST_VALUE,  # ARRAY_CONTAINS_ANY_OF_HASHED
    STRING_LIST_VALUE,  # ARRAY_NOT_CONTAINS_ANY_OF_HASHED
    STRING_VALUE,       # EQUALS
    STRING_VALUE,       # NOT_EQUALS
    STRING_LIST_VALUE,  # STARTS_WITH_ANY_OF
    STRING_LIST_VALUE,  # NOT_STARTS_WITH_ANY_OF
    STRING_LIST_VALUE,  # ENDS_WITH_ANY_OF
    STRING_LIST_VALUE,  # NOT_ENDS_WITH_ANY_OF
    STRING_LIST_VALUE,  # ARRAY_CONTAINS_ANY_OF
    STRING_LIST_VALUE   # ARRAY_NOT_CONTAINS_ANY_OF
]
SEGMENT_COMPARATOR_TEXTS = ['IS IN SEGMENT', 'IS NOT IN SEGMENT']
PREREQUISITE_COMPARATOR_TEXTS = ['EQUALS', 'DOES NOT EQUAL']


def fixup_config_salt_and_segments(config):
    """
    Adds the inline salt and segment to the config.
    When using flag overrides, the original salt and segment indexes may become invalid. Therefore, we copy the
    object references to the locations where they are referenced and use these references instead of the indexes.
    """
    salt = config.get(PREFERENCES, {}).get(SALT, '')
    segments = config.get(SEGMENTS, [])
    settings = config.get(FEATURE_FLAGS, {})
    for setting in settings.values():
        if not isinstance(setting, dict):
            continue

        # add salt
        setting[INLINE_SALT] = salt

        # add segment to the segment conditions
        targeting_rules = setting.get(TARGETING_RULES, [])
        for targeting_rule in targeting_rules:
            conditions = targeting_rule.get(CONDITIONS, [])
            for condition in conditions:

                segment_condition = condition.get(SEGMENT_CONDITION)
                if segment_condition:
                    segment_index = segment_condition.get(SEGMENT_INDEX)
                    segment = segments[segment_index]
                    segment_condition[INLINE_SEGMENT] = segment



import json
import sys
from math import floor

from . import utils
from .config import fixup_config_salt_and_segments
from .utils import unicode_to_utf8


class ConfigEntry(object):
    def __init__(self, config=None, etag='', config_json_string='{}', fetch_time=utils.distant_past):
        self.config = config if config is not None else {}
        self.etag = etag
        self.config_json_string = config_json_string
        self.fetch_time = fetch_time

    def is_empty(self):
        return self == ConfigEntry.empty

    def serialize(self):
        return '{:.0f}\n{}\n{}'.format(floor(self.fetch_time * 1000), self.etag, self.config_json_string)

    @classmethod
    def create_from_string(cls, string):
        if not string:
            return ConfigEntry.empty

        fetch_time_index = string.find('\n')
        etag_index = string.find('\n', fetch_time_index + 1)
        if fetch_time_index < 0 or etag_index < 0:
            raise ValueError('Number of values is fewer than expected.')

        try:
            fetch_time = float(string[0:fetch_time_index])
        except ValueError:
            raise ValueError('Invalid fetch time: {}'.format(string[0:fetch_time_index]))

        etag = string[fetch_time_index + 1:etag_index]
        if not etag:
            raise ValueError('Empty eTag value')
        try:
            config_json = string[etag_index + 1:]
            config = json.loads(config_json)
            if sys.version_info[0] == 2:
                config = unicode_to_utf8(config)  # On Python 2.7, convert unicode to utf-8
            fixup_config_salt_and_segments(config)
        except ValueError as e:
            raise ValueError('Invalid config JSON: {}. {}'.format(config_json, str(e)))

        return ConfigEntry(config=config, etag=etag, config_json_string=config_json, fetch_time=fetch_time / 1000.0)


ConfigEntry.empty = ConfigEntry(etag='empty')



from abc import ABCMeta, abstractmethod


class ConfigCache(object):
    """
    Config cache interface
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def get(self, key):
        """
        :return: the config json object from the cache
        """

    @abstractmethod
    def set(self, key, value):
        """
        Sets the config json cache.
        """


class ConfigCatClientException(Exception):
    """
    Generic ConfigCatClientException
    """



class EvaluationContext(object):
    def __init__(self,
                 key,
                 setting_type,
                 user,
                 visited_keys=None,
                 is_missing_user_object_logged=False,
                 is_missing_user_object_attribute_logged=False):
        self.key = key
        self.setting_type = setting_type
        self.user = user
        self.visited_keys = visited_keys
        self.is_missing_user_object_logged = is_missing_user_object_logged
        self.is_missing_user_object_attribute_logged = is_missing_user_object_attribute_logged



class EvaluationDetails(object):
    def __init__(self,
                 key,
                 value,
                 variation_id=None,
                 fetch_time=None,
                 user=None,
                 is_default_value=False,
                 error=None,
                 matched_targeting_rule=None,
                 matched_percentage_option=None):
        # Key of the feature flag or setting.
        self.key = key

        # Evaluated value of the feature flag or setting.
        self.value = value

        # Variation ID of the feature flag or setting (if available).
        self.variation_id = variation_id

        # Time of last successful config download.
        self.fetch_time = fetch_time

        # The User Object used for the evaluation (if available).
        self.user = user

        # Indicates whether the default value passed to the setting evaluation methods like ConfigCatClient.get_value,
        # ConfigCatClient.get_value_details, etc. is used as the result of the evaluation.
        self.is_default_value = is_default_value

        # Error message in case evaluation failed.
        self.error = error

        # The targeting rule (if any) that matched during the evaluation and was used to return the evaluated value.
        self.matched_targeting_rule = matched_targeting_rule

        # The percentage option (if any) that was used to select the evaluated value.
        self.matched_percentage_option = matched_percentage_option

    @staticmethod
    def from_error(key, value, error, variation_id=None):
        return EvaluationDetails(key=key, value=value, variation_id=variation_id, is_default_value=True, error=error)



import hashlib
from threading import Thread, Event, Lock

from . import utils
from .config import FEATURE_FLAGS, CONFIG_FILE_NAME, SERIALIZATION_FORMAT_VERSION
from .configentry import ConfigEntry
from .pollingmode import AutoPollingMode, LazyLoadingMode
from .refreshresult import RefreshResult


class ConfigService(object):
    def __init__(self, sdk_key, polling_mode, hooks, config_fetcher, log, config_cache, is_offline):
        self._cached_entry = ConfigEntry.empty
        self._cached_entry_string = ''
        self._polling_mode = polling_mode
        self.log = log
        self._config_cache = config_cache
        self._hooks = hooks
        self._cache_key = ConfigService._get_cache_key(sdk_key)
        self._config_fetcher = config_fetcher
        self._is_offline = is_offline
        self._response_future = None
        self._initialized = Event()
        self._lock = Lock()
        self._ongoing_fetch = False
        self._fetch_finished = Event()
        self._start_time = utils.get_utc_now()

        if isinstance(self._polling_mode, AutoPollingMode) and not is_offline:
            self._start_poll()
        else:
            self._set_initialized()

    def get_config(self):
        if isinstance(self._polling_mode, LazyLoadingMode):
            entry, _ = self._fetch_if_older(
                utils.get_utc_now_seconds_since_epoch() - self._polling_mode.cache_refresh_interval_seconds)
            return (entry.config, entry.fetch_time) \
                if not entry.is_empty() \
                else (None, utils.distant_past)
        elif isinstance(self._polling_mode, AutoPollingMode) and not self._initialized.is_set():
            elapsed_time = (utils.get_utc_now() - self._start_time).total_seconds()
            if elapsed_time < self._polling_mode.max_init_wait_time_seconds:
                self._initialized.wait(self._polling_mode.max_init_wait_time_seconds - elapsed_time)

                # Max wait time expired without result, notify subscribers with the cached config.
                if not self._initialized.is_set():
                    self._set_initialized()
                    return (self._cached_entry.config, self._cached_entry.fetch_time) \
                        if not self._cached_entry.is_empty() \
                        else (None, utils.distant_past)

        # If we are initialized, we prefer the cached results
        entry, _ = self._fetch_if_older(utils.distant_past, prefer_cache=self._initialized.is_set())
        return (entry.config, entry.fetch_time) \
            if not entry.is_empty() \
            else (None, utils.distant_past)

    def refresh(self):
        """
        :return: RefreshResult object
        """
        if self.is_offline():
            offline_warning = 'Client is in offline mode, it cannot initiate HTTP calls.'
            self.log.warning(offline_warning, event_id=3200)
            return RefreshResult(is_success=False, error=offline_warning)

        _, error = self._fetch_if_older(utils.distant_future)
        return RefreshResult(is_success=error is None, error=error)

    def set_online(self):
        with self._lock:
            if not self._is_offline:
                return

            self._is_offline = False
            if isinstance(self._polling_mode, AutoPollingMode):
                self._start_poll()

            self.log.info('Switched to %s mode.', 'ONLINE', event_id=5200)

    def set_offline(self):
        with self._lock:
            if self._is_offline:
                return

            self._is_offline = True
            if isinstance(self._polling_mode, AutoPollingMode):
                self._stopped.set()
                self._thread.join()

            self.log.info('Switched to %s mode.', 'OFFLINE', event_id=5200)

    def is_offline(self):
        return self._is_offline  # atomic operation in python (lock is not needed)

    def close(self):
        if isinstance(self._polling_mode, AutoPollingMode):
            self._stopped.set()

    def _fetch_if_older(self, threshold, prefer_cache=False):
        """
        :return: Returns the ConfigEntry object and error message in case of any error.
        """

        with self._lock:
            # Sync up with the cache and use it when it's not expired.
            from_cache = self._read_cache()
            if not from_cache.is_empty() and from_cache.etag != self._cached_entry.etag:
                self._cached_entry = from_cache
                self._hooks.invoke_on_config_changed(from_cache.config.get(FEATURE_FLAGS))

            # Cache isn't expired
            if self._cached_entry.fetch_time > threshold:
                self._set_initialized()
                return self._cached_entry, None

            # If we are in offline mode or the caller prefers cached values, do not initiate fetch.
            if self._is_offline or prefer_cache:
                return self._cached_entry, None

        # No fetch is running, initiate a new one.
        # Ensure only one fetch request is running at a time.
        # If there's an ongoing fetch running, we will wait for the ongoing fetch.
        if self._ongoing_fetch:
            self._fetch_finished.wait()
        else:
            self._ongoing_fetch = True
            self._fetch_finished.clear()
            response = self._config_fetcher.get_configuration(self._cached_entry.etag)

            with self._lock:
                if response.is_fetched():
                    self._cached_entry = response.entry
                    self._write_cache(response.entry)
                    self._hooks.invoke_on_config_changed(response.entry.config.get(FEATURE_FLAGS))
                elif (response.is_not_modified() or not response.is_transient_error) and \
                        not self._cached_entry.is_empty():
                    self._cached_entry.fetch_time = utils.get_utc_now_seconds_since_epoch()
                    self._write_cache(self._cached_entry)

                self._set_initialized()

            self._ongoing_fetch = False
            self._fetch_finished.set()

        return self._cached_entry, None

    def _start_poll(self):
        self._started = Event()
        self._thread = Thread(target=self._run, args=[])
        self._thread.daemon = True  # daemon thread terminates its execution when the main thread terminates
        self._thread.start()
        self._started.wait()

    def _run(self):
        self._stopped = Event()
        self._started.set()
        while True:
            self._fetch_if_older(utils.get_utc_now_seconds_since_epoch() - self._polling_mode.poll_interval_seconds)
            self._stopped.wait(timeout=self._polling_mode.poll_interval_seconds)
            if self._stopped.is_set():
                break

    def _set_initialized(self):
        if not self._initialized.is_set():
            self._initialized.set()
            self._hooks.invoke_on_client_ready()

    @staticmethod
    def _get_cache_key(sdk_key):
        return hashlib.sha1(
            (sdk_key + '_' + CONFIG_FILE_NAME + '.json' + '_' + SERIALIZATION_FORMAT_VERSION).encode('utf-8')).hexdigest()

    def _read_cache(self):
        try:
            json_string = self._config_cache.get(self._cache_key)
            if not json_string or json_string == self._cached_entry_string:
                return ConfigEntry.empty

            self._cached_entry_string = json_string
            return ConfigEntry.create_from_string(json_string)
        except Exception:
            self.log.exception('Error occurred while reading the cache.', event_id=2200)
            return ConfigEntry.empty

    def _write_cache(self, config_entry):
        try:
            self._config_cache.set(self._cache_key, config_entry.serialize())
        except Exception:
            self.log.exception('Error occurred while writing the cache.', event_id=2201)



import json

import hashlib
import math
import sys
import semver

from .config import FEATURE_FLAGS, INLINE_SALT, TARGETING_RULES, PERCENTAGE_RULE_ATTRIBUTE, CONDITIONS, SERVED_VALUE, \
    get_value, VARIATION_ID, TARGETING_RULE_PERCENTAGE_OPTIONS, PERCENTAGE_OPTIONS, PERCENTAGE, USER_CONDITION, \
    SEGMENT_CONDITION, PREREQUISITE_FLAG_CONDITION, PREREQUISITE_FLAG_KEY, PREREQUISITE_COMPARATOR, \
    PrerequisiteComparator, INLINE_SEGMENT, SEGMENT_NAME, SEGMENT_COMPARATOR, SEGMENT_CONDITIONS, SegmentComparator, \
    COMPARISON_ATTRIBUTE, COMPARATOR, Comparator, COMPARATOR_TEXTS, PREREQUISITE_COMPARATOR_TEXTS, \
    SEGMENT_COMPARATOR_TEXTS, COMPARISON_VALUES, get_value_type, SETTING_TYPE, SettingType
from .evaluationcontext import EvaluationContext
from .evaluationlogbuilder import EvaluationLogBuilder
from .logger import Logger
from datetime import datetime

from .user import User
from .utils import unicode_to_utf8, encode_utf8, get_seconds_since_epoch, is_string_list


def sha256(value_utf8, salt, context_salt):
    """
    Calculates the SHA256 hash of the given value with the given salt and context_salt.
    """
    return hashlib.sha256(value_utf8 + salt.encode('utf-8') + context_salt.encode('utf-8')).hexdigest()


class RolloutEvaluator(object):
    def __init__(self, log):
        self.log = log

    def evaluate(self, key, user, default_value, default_variation_id, config, log_builder, visited_keys=None):  # noqa: C901
        """
        returns value, variation_id, matched_targeting_rule, matched_percentage_option, error
        """

        if visited_keys is None:
            visited_keys = []
        is_root_flag_evaluation = len(visited_keys) == 0

        settings = config.get(FEATURE_FLAGS, {})
        setting_descriptor = settings.get(key)

        if setting_descriptor is None:
            error = 'Failed to evaluate setting \'%s\' (the key was not found in config JSON). ' \
                    'Returning the `%s` parameter that you specified in your application: \'%s\'. ' \
                    'Available keys: [%s].'
            error_args = (key, 'default_value', str(default_value), ', '.join("'{}'".format(s) for s in list(settings)))
            self.log.error(error, *error_args, event_id=1001)
            return default_value, default_variation_id, None, None, Logger.format(error, error_args)

        setting_type = setting_descriptor.get(SETTING_TYPE)
        salt = setting_descriptor.get(INLINE_SALT, '')
        targeting_rules = setting_descriptor.get(TARGETING_RULES, [])
        percentage_rule_attribute = setting_descriptor.get(PERCENTAGE_RULE_ATTRIBUTE)

        context = EvaluationContext(key, setting_type, user, visited_keys)

        user_has_invalid_type = context.user is not None and not isinstance(context.user, User)
        if user_has_invalid_type:
            self.log.warning('Cannot evaluate targeting rules and %% options for setting \'%s\' '
                             '(User Object is not an instance of User type). '
                             'You should pass a User Object to the evaluation methods like `get_value()` '
                             'in order to make targeting work properly. '
                             'Read more: https://configcat.com/docs/advanced/user-object/',
                             key, event_id=4001)
            # We set the user to None and won't log further missing user object warnings
            context.user = None
            context.is_missing_user_object_logged = True

        try:
            if log_builder and is_root_flag_evaluation:
                log_builder.append("Evaluating '{}'".format(key))
                log_builder.append(" for User '{}'".format(context.user) if context.user is not None else '')
                log_builder.increase_indent()

            # Evaluate targeting rules (logically connected by OR)
            if log_builder and len(targeting_rules) > 0:
                log_builder.new_line('Evaluating targeting rules and applying the first match if any:')
            for targeting_rule in targeting_rules:
                conditions = targeting_rule.get(CONDITIONS, [])

                if len(conditions) > 0:
                    served_value = targeting_rule.get(SERVED_VALUE)
                    value = get_value(served_value, setting_type) if served_value is not None else None

                    # Evaluate targeting rule conditions (logically connected by AND)
                    if self._evaluate_conditions(conditions, context, salt, config, log_builder, value):
                        if served_value is not None:
                            variation_id = served_value.get(VARIATION_ID, default_variation_id)
                            log_builder and is_root_flag_evaluation and log_builder.new_line("Returning '%s'." % value)
                            return value, variation_id, targeting_rule, None, None
                    else:
                        continue

                # Evaluate percentage options of the targeting rule
                log_builder and log_builder.increase_indent()
                percentage_options = targeting_rule.get(TARGETING_RULE_PERCENTAGE_OPTIONS, [])
                percentage_evaluation_result, percentage_value, percentage_variation_id, percentage_option = \
                    self._evaluate_percentage_options(percentage_options, context, percentage_rule_attribute,
                                                      default_variation_id, log_builder)

                if percentage_evaluation_result:
                    if log_builder:
                        log_builder.decrease_indent()
                        is_root_flag_evaluation and log_builder.new_line("Returning '%s'." % percentage_value)
                    return percentage_value, percentage_variation_id, targeting_rule, percentage_option, None
                else:
                    if log_builder:
                        log_builder.new_line(
                            'The current targeting rule is ignored and the evaluation continues with the next rule.')
                        log_builder.decrease_indent()
                    continue

            # Evaluate percentage options
            percentage_options = setting_descriptor.get(PERCENTAGE_OPTIONS, [])
            percentage_evaluation_result, percentage_value, percentage_variation_id, percentage_option = \
                self._evaluate_percentage_options(percentage_options, context, percentage_rule_attribute,
                                                  default_variation_id, log_builder)
            if percentage_evaluation_result:
                log_builder and is_root_flag_evaluation and log_builder.new_line("Returning '%s'." % percentage_value)
                return percentage_value, percentage_variation_id, None, percentage_option, None

            return_value = get_value(setting_descriptor, setting_type)
            return_variation_id = setting_descriptor.get(VARIATION_ID, default_variation_id)
            log_builder and is_root_flag_evaluation and log_builder.new_line("Returning '%s'." % return_value)
            return return_value, return_variation_id, None, None, None
        except Exception as e:
            # During the recursive evaluation of a prerequisite flag, we propagate the exceptions
            # and let the root flag's evaluation code handle them.
            if not is_root_flag_evaluation:
                raise e

            error = 'Failed to evaluate setting \'%s\'. (%s). ' \
                    'Returning the `%s` parameter that you specified in your application: \'%s\'. '
            error_args = (key, str(e), 'default_value', str(default_value))
            self.log.error(error, *error_args, event_id=2001)
            return default_value, default_variation_id, None, None, Logger.format(error, error_args)

    def _format_rule(self, comparison_attribute, comparator, comparison_value):
        comparator_text = COMPARATOR_TEXTS[comparator]
        return "User.%s %s %s" \
               % (comparison_attribute, comparator_text,
                  EvaluationLogBuilder.trunc_comparison_value_if_needed(comparator, comparison_value))

    def _user_attribute_value_to_string(self, value):
        if value is None:
            return None

        if isinstance(value, datetime):
            value = self._get_user_attribute_value_as_seconds_since_epoch(value)
        elif is_string_list(value):
            value = self._get_user_attribute_value_as_string_list(value)
            return json.dumps(value, ensure_ascii=False, separators=(',', ':'))  # Convert the list to a JSON string

        if isinstance(value, float):
            if math.isnan(value):
                return 'NaN'
            if value == float('inf'):
                return 'Infinity'
            if value == float('-inf'):
                return '-Infinity'
            if 'e' in str(value):
                return str(value)
            if value.is_integer():
                return str(int(value))

        return str(value)

    def _get_user_attribute_value_as_text(self, attribute_name, attribute_value, condition, key):
        if isinstance(attribute_value, str):
            return attribute_value

        if sys.version_info[0] == 2 and isinstance(attribute_value, unicode):  # noqa: F821
            return attribute_value  # Handle unicode strings on Python 2.7

        self.log.warning('Evaluation of condition (%s) for setting \'%s\' may not produce the expected result '
                         '(the User.%s attribute is not a string value, thus it was automatically converted to '
                         'the string value \'%s\'). Please make sure that using a non-string value was intended.',
                         condition, key, attribute_name, attribute_value, event_id=3005)
        return self._user_attribute_value_to_string(attribute_value)

    def _convert_numeric_to_float(self, value):
        if isinstance(value, str):
            return float(value.replace(",", "."))

        if sys.version_info[0] == 2 and isinstance(value, unicode):  # noqa: F821
            return float(value.replace(",", "."))  # Handle unicode strings on Python 2.7

        return float(value)

    def _get_user_attribute_value_as_seconds_since_epoch(self, attribute_value):
        if isinstance(attribute_value, datetime):
            return get_seconds_since_epoch(attribute_value)

        return self._convert_numeric_to_float(attribute_value)

    def _get_user_attribute_value_as_string_list(self, attribute_value):
        # Handle unicode strings on Python 2.7
        if isinstance(attribute_value, str) or sys.version_info[0] == 2 and isinstance(attribute_value, unicode):  # noqa: F821
            attribute_value_list = json.loads(attribute_value)
        else:
            attribute_value_list = attribute_value

        if not is_string_list(attribute_value_list):
            raise ValueError()

        return attribute_value_list

    def _handle_invalid_user_attribute(self, comparison_attribute, comparator, comparison_value, key, validation_error):
        """
        returns: evaluation error message
        """
        error = 'cannot evaluate, the User.%s attribute is invalid (%s)' % (comparison_attribute, validation_error)
        self.log.warning('Cannot evaluate condition (%s) for setting \'%s\' '
                         '(%s). Please check the User.%s attribute and make sure that its value corresponds to the '
                         'comparison operator.',
                         self._format_rule(comparison_attribute, comparator, comparison_value), key,
                         validation_error, comparison_attribute, event_id=3004)
        return error

    def _evaluate_percentage_options(self, percentage_options, context, percentage_rule_attribute, default_variation_id, log_builder):  # noqa: C901, E501
        """
        returns: evaluation_result, percentage_value, percentage_variation_id, percentage_option
        """
        if len(percentage_options) == 0:
            return False, None, None, None

        user = context.user
        key = context.key

        if user is None:
            if not context.is_missing_user_object_logged:
                self.log.warning('Cannot evaluate targeting rules and %% options for setting \'%s\' '
                                 '(User Object is missing). '
                                 'You should pass a User Object to the evaluation methods like `get_value()` '
                                 'in order to make targeting work properly. '
                                 'Read more: https://configcat.com/docs/advanced/user-object/',
                                 key, event_id=3001)
                context.is_missing_user_object_logged = True

            if log_builder:
                log_builder.new_line('Skipping % options because the User Object is missing.')
            return False, None, None, None

        user_attribute_name = percentage_rule_attribute if percentage_rule_attribute is not None else 'Identifier'
        if percentage_rule_attribute is not None:
            user_key = user.get_attribute(percentage_rule_attribute)
        else:
            user_key = user.get_identifier()
        if percentage_rule_attribute is not None and user_key is None:
            if not context.is_missing_user_object_attribute_logged:
                self.log.warning('Cannot evaluate %% options for setting \'%s\' '
                                 '(the User.%s attribute is missing). You should set the User.%s attribute in order to make '
                                 'targeting work properly. Read more: https://configcat.com/docs/advanced/user-object/',
                                 key, percentage_rule_attribute, percentage_rule_attribute,
                                 event_id=3003)
                context.is_missing_user_object_attribute_logged = True

            if log_builder:
                log_builder.new_line(
                    'Skipping %% options because the User.%s attribute is missing.' % user_attribute_name)
            return False, None, None, None

        # Unicode fix on Python 2.7
        if sys.version_info[0] == 2:
            try:
                hash_candidate = ('%s%s' % (key, self._user_attribute_value_to_string(user_key))).encode('utf-8')
            except Exception:
                hash_candidate = ('%s%s' % (key, self._user_attribute_value_to_string(user_key))).decode('utf-8').encode(
                    'utf-8')
        else:
            hash_candidate = ('%s%s' % (key, self._user_attribute_value_to_string(user_key))).encode('utf-8')
        hash_val = int(hashlib.sha1(hash_candidate).hexdigest()[:7], 16) % 100

        if log_builder:
            log_builder.new_line('Evaluating %% options based on the User.%s attribute:' % user_attribute_name)
            log_builder.new_line('- Computing hash in the [0..99] range from User.%s => %s '
                                 '(this value is sticky and consistent across all SDKs)' %
                                 (user_attribute_name, hash_val))

        bucket = 0
        index = 1
        for percentage_option in percentage_options or []:
            percentage = percentage_option.get(PERCENTAGE, 0)
            bucket += percentage
            if hash_val < bucket:
                percentage_value = get_value(percentage_option, context.setting_type)
                variation_id = percentage_option.get(VARIATION_ID, default_variation_id)
                if log_builder:
                    log_builder.new_line("- Hash value %s selects %% option %s (%s%%), '%s'." %
                                         (hash_val, index, percentage, percentage_value))
                return True, percentage_value, variation_id, percentage_option
            index += 1

        return False, None, None, None

    def _evaluate_conditions(self, conditions, context, salt, config, log_builder, value):  # noqa: C901
        first_condition = True
        condition_result = True
        error = None
        for condition in conditions:
            user_condition = condition.get(USER_CONDITION)
            segment_condition = condition.get(SEGMENT_CONDITION)
            prerequisite_flag_condition = condition.get(PREREQUISITE_FLAG_CONDITION)

            if first_condition:
                if log_builder:
                    log_builder.new_line('- IF ')
                    log_builder.increase_indent()
                first_condition = False
            else:
                if log_builder:
                    log_builder.new_line('AND ')

            if user_condition is not None:
                result, error = self._evaluate_user_condition(user_condition, context, context.key, salt,
                                                              log_builder)
                if log_builder and len(conditions) > 1:
                    log_builder.append('=> {}'.format('true' if result else 'false'))
                    if not result:
                        log_builder.append(', skipping the remaining AND conditions')

                if not result or error:
                    condition_result = False
                    break
            elif segment_condition is not None:
                result, error = self._evaluate_segment_condition(segment_condition, context, salt, log_builder)
                if log_builder:
                    if len(conditions) > 1:
                        if error is None:
                            log_builder.append(' ')
                        log_builder.append('=> {}'.format('true' if result else 'false'))
                        if not result:
                            log_builder.append(', skipping the remaining AND conditions')
                    elif error is None:
                        log_builder.new_line()

                if not result or error:
                    condition_result = False
                    break
            elif prerequisite_flag_condition is not None:
                result = self._evaluate_prerequisite_flag_condition(prerequisite_flag_condition, context, config, log_builder)
                if log_builder:
                    if len(conditions) > 1:
                        log_builder.append(' => {}'.format('true' if result else 'false'))
                        if not result:
                            log_builder.append(', skipping the remaining AND conditions')
                    elif error is None:
                        log_builder.new_line()

                if not result:
                    condition_result = False
                    break

        if log_builder:
            if len(conditions) > 1:
                log_builder.new_line()
            if error:
                log_builder.append('THEN %s => %s' % (
                    "'" + str(value) + "'" if value is not None else '% options',
                    error))
                log_builder.new_line(
                    'The current targeting rule is ignored and the evaluation continues with the next rule.')
            else:
                log_builder.append('THEN %s => %s' % (
                    "'" + str(value) + "'" if value is not None else '% options',
                    'MATCH, applying rule' if condition_result else 'no match'))
            if len(conditions) > 0:
                log_builder.decrease_indent()

        return condition_result

    def _evaluate_prerequisite_flag_condition(self, prerequisite_flag_condition, context, config, log_builder):  # noqa: C901, E501
        prerequisite_key = prerequisite_flag_condition.get(PREREQUISITE_FLAG_KEY)
        prerequisite_comparator = prerequisite_flag_condition.get(PREREQUISITE_COMPARATOR)

        # Check if the prerequisite key exists
        settings = config.get(FEATURE_FLAGS, {})
        if prerequisite_key is None or settings.get(prerequisite_key) is None:
            raise ValueError('Prerequisite flag key is missing or invalid.')

        prerequisite_condition_result = False
        prerequisite_flag_setting_type = settings[prerequisite_key].get(SETTING_TYPE)
        prerequisite_comparison_value_type = get_value_type(prerequisite_flag_condition)

        prerequisite_comparison_value = get_value(prerequisite_flag_condition, prerequisite_flag_setting_type)

        # Type mismatch check
        if prerequisite_comparison_value_type != SettingType.to_type(prerequisite_flag_setting_type):
            raise ValueError("Type mismatch between comparison value '%s' and prerequisite flag '%s'" %
                             (prerequisite_comparison_value, prerequisite_key))

        prerequisite_condition = ("Flag '%s' %s '%s'" %
                                  (prerequisite_key, PREREQUISITE_COMPARATOR_TEXTS[prerequisite_comparator],
                                   str(prerequisite_comparison_value)))

        # Circular dependency check
        visited_keys = context.visited_keys
        visited_keys.append(context.key)
        if prerequisite_key in visited_keys:
            depending_flags = ' -> '.join("'{}'".format(s) for s in list(visited_keys) + [prerequisite_key])
            raise ValueError('Circular dependency detected between the following depending flags: %s.' % depending_flags)

        if log_builder:
            log_builder.append(prerequisite_condition)
            log_builder.new_line('(').increase_indent()
            log_builder.new_line("Evaluating prerequisite flag '%s':" % prerequisite_key)

        prerequisite_value, _, _, _, _ = self.evaluate(prerequisite_key, context.user, None, None, config,
                                                       log_builder, context.visited_keys)

        visited_keys.pop()

        if log_builder:
            log_builder.new_line("Prerequisite flag evaluation result: '%s'." % str(prerequisite_value))
            log_builder.new_line("Condition (Flag '%s' %s '%s') evaluates to " %
                                 (prerequisite_key, PREREQUISITE_COMPARATOR_TEXTS[prerequisite_comparator],
                                  str(prerequisite_comparison_value)))

        # EQUALS
        if prerequisite_comparator == PrerequisiteComparator.EQUALS:
            if prerequisite_value == prerequisite_comparison_value:
                prerequisite_condition_result = True
        # DOES NOT EQUAL
        elif prerequisite_comparator == PrerequisiteComparator.NOT_EQUALS:
            if prerequisite_value != prerequisite_comparison_value:
                prerequisite_condition_result = True
        else:
            raise ValueError('Comparison operator is missing or invalid.')

        if log_builder:
            log_builder.append('%s.' % ('true' if prerequisite_condition_result else 'false'))
            log_builder.decrease_indent().new_line(')')

        return prerequisite_condition_result

    def _evaluate_segment_condition(self, segment_condition, context, salt, log_builder):  # noqa: C901
        user = context.user
        key = context.key

        segment = segment_condition[INLINE_SEGMENT]
        if segment is None:
            raise ValueError('Segment reference is invalid.')

        segment_name = segment.get(SEGMENT_NAME, '')
        segment_comparator = segment_condition.get(SEGMENT_COMPARATOR)
        segment_conditions = segment.get(SEGMENT_CONDITIONS, [])

        if user is None:
            if not context.is_missing_user_object_logged:
                self.log.warning('Cannot evaluate targeting rules and %% options for setting \'%s\' '
                                 '(User Object is missing). '
                                 'You should pass a User Object to the evaluation methods like `get_value()` '
                                 'in order to make targeting work properly. '
                                 'Read more: https://configcat.com/docs/advanced/user-object/',
                                 key, event_id=3001)
                context.is_missing_user_object_logged = True
            if log_builder:
                log_builder.append("User %s '%s' " % (SEGMENT_COMPARATOR_TEXTS[segment_comparator], segment_name))
            return False, 'cannot evaluate, User Object is missing'

        # IS IN SEGMENT, IS NOT IN SEGMENT
        if segment_comparator in [SegmentComparator.IS_IN, SegmentComparator.IS_NOT_IN]:
            if log_builder:
                log_builder.append("User %s '%s'" %
                                   (SEGMENT_COMPARATOR_TEXTS[segment_comparator], segment_name))
                log_builder.new_line('(').increase_indent()
                log_builder.new_line("Evaluating segment '%s':" % segment_name)

            # Set initial condition result based on comparator
            segment_condition_result = segment_comparator == SegmentComparator.IS_IN

            # Evaluate segment conditions (logically connected by AND)
            first_segment_rule = True
            error = None
            for segment_condition in segment_conditions:
                if first_segment_rule:
                    if log_builder:
                        log_builder.new_line('- IF ')
                        log_builder.increase_indent()
                    first_segment_rule = False
                else:
                    if log_builder:
                        log_builder.new_line('AND ')

                result, error = self._evaluate_user_condition(segment_condition, context, segment_name, salt, log_builder)
                if log_builder:
                    log_builder.append('=> {}'.format('true' if result else 'false'))
                    if not result:
                        log_builder.append(', skipping the remaining AND conditions')

                if not result:
                    segment_condition_result = False if segment_comparator == SegmentComparator.IS_IN else True
                    break

            if log_builder:
                log_builder.decrease_indent()
                segment_evaluation_result = segment_condition_result if segment_comparator == SegmentComparator.IS_IN \
                    else not segment_condition_result
                log_builder.new_line('Segment evaluation result: ')
                if not error:
                    log_builder.append('User IS%sIN SEGMENT.' % (' ' if segment_evaluation_result else ' NOT '))
                else:
                    log_builder.append('%s.' % error)

                log_builder.new_line("Condition (User %s '%s') " % (SEGMENT_COMPARATOR_TEXTS[segment_comparator],
                                                                    segment_name))
                if not error:
                    log_builder.append("evaluates to %s." % ('true' if segment_condition_result else 'false'))
                else:
                    log_builder.append('failed to evaluate.')

                log_builder.decrease_indent().new_line(')')
                if error:
                    log_builder.new_line()

            return segment_condition_result, error

        raise ValueError('Comparison operator is missing or invalid.')

    def _evaluate_user_condition(self, user_condition, context, context_salt, salt, log_builder):  # noqa: C901, E501
        """
        returns result of user condition, error
        """

        user = context.user
        key = context.key

        comparison_attribute = user_condition.get(COMPARISON_ATTRIBUTE)
        comparator = user_condition.get(COMPARATOR)
        comparison_value = user_condition.get(COMPARISON_VALUES[comparator])
        condition = self._format_rule(comparison_attribute, comparator, comparison_value)
        error = None

        if comparison_attribute is None:
            raise ValueError('Comparison attribute name is missing.')

        if log_builder:
            log_builder.append(condition + ' ')

        if user is None:
            if not context.is_missing_user_object_logged:
                self.log.warning('Cannot evaluate targeting rules and %% options for setting \'%s\' '
                                 '(User Object is missing). '
                                 'You should pass a User Object to the evaluation methods like `get_value()` '
                                 'in order to make targeting work properly. '
                                 'Read more: https://configcat.com/docs/advanced/user-object/',
                                 key, event_id=3001)
                context.is_missing_user_object_logged = True
            error = 'cannot evaluate, User Object is missing'
            return False, error

        user_value = user.get_attribute(comparison_attribute)
        if user_value is None or (isinstance(user_value, str) and len(user_value) == 0):
            self.log.warning('Cannot evaluate condition (%s) for setting \'%s\' '
                             '(the User.%s attribute is missing). You should set the User.%s attribute in order to make '
                             'targeting work properly. Read more: https://configcat.com/docs/advanced/user-object/',
                             condition, key, comparison_attribute, comparison_attribute, event_id=3003)
            error = 'cannot evaluate, the User.{} attribute is missing'.format(comparison_attribute)
            return False, error

        # IS ONE OF
        if comparator == Comparator.IS_ONE_OF:
            user_value = self._get_user_attribute_value_as_text(comparison_attribute, user_value, condition, key)
            if user_value in comparison_value:
                return True, error
        # IS NOT ONE OF
        elif comparator == Comparator.IS_NOT_ONE_OF:
            user_value = self._get_user_attribute_value_as_text(comparison_attribute, user_value, condition, key)
            if user_value not in comparison_value:
                return True, error
        # CONTAINS ANY OF
        elif comparator == Comparator.CONTAINS_ANY_OF:
            user_value = self._get_user_attribute_value_as_text(comparison_attribute, user_value, condition, key)
            for comparison in comparison_value:
                if comparison in user_value:
                    return True, error
        # NOT CONTAINS ANY OF
        elif comparator == Comparator.NOT_CONTAINS_ANY_OF:
            user_value = self._get_user_attribute_value_as_text(comparison_attribute, user_value, condition, key)
            if not any(comparison in user_value for comparison in comparison_value):
                return True, error
        # IS ONE OF, IS NOT ONE OF (Semantic version)
        elif Comparator.IS_ONE_OF_SEMVER <= comparator <= Comparator.IS_NOT_ONE_OF_SEMVER:
            try:
                match = False
                for x in filter(None, [x.strip() for x in comparison_value]):
                    match = semver.VersionInfo.parse(str(user_value).strip()).match('==' + str(x)) or match
                if match == (comparator == Comparator.IS_ONE_OF_SEMVER):
                    return True, error
            except ValueError:
                validation_error = "'%s' is not a valid semantic version" % str(user_value).strip()
                error = self._handle_invalid_user_attribute(comparison_attribute, comparator, comparison_value, key,
                                                            validation_error)
                return False, error
        # LESS THAN, LESS THAN OR EQUALS TO, GREATER THAN, GREATER THAN OR EQUALS TO (Semantic version)
        elif Comparator.LESS_THAN_SEMVER <= comparator <= Comparator.GREATER_THAN_OR_EQUAL_SEMVER:
            try:
                if semver.VersionInfo.parse(str(user_value).strip()).match(
                        COMPARATOR_TEXTS[comparator] + str(comparison_value).strip()
                ):
                    return True, error
            except ValueError:
                validation_error = "'%s' is not a valid semantic version" % str(user_value).strip()
                error = self._handle_invalid_user_attribute(comparison_attribute, comparator, comparison_value, key,
                                                            validation_error)
                return False, error
        # =, <>, <, <=, >, >= (number)
        elif Comparator.EQUALS_NUMBER <= comparator <= Comparator.GREATER_THAN_OR_EQUAL_NUMBER:
            try:
                user_value_float = self._convert_numeric_to_float(user_value)
            except ValueError:
                validation_error = "'%s' is not a valid decimal number" % str(user_value)
                error = self._handle_invalid_user_attribute(comparison_attribute, comparator, comparison_value, key,
                                                            validation_error)
                return False, error

            comparison_value_float = float(comparison_value)

            if (comparator == Comparator.EQUALS_NUMBER and user_value_float == comparison_value_float) \
                    or (comparator == Comparator.NOT_EQUALS_NUMBER and user_value_float != comparison_value_float) \
                    or (comparator == Comparator.LESS_THAN_NUMBER and user_value_float < comparison_value_float) \
                    or (comparator == Comparator.LESS_THAN_OR_EQUAL_NUMBER and user_value_float <= comparison_value_float) \
                    or (comparator == Comparator.GREATER_THAN_NUMBER and user_value_float > comparison_value_float) \
                    or (comparator == Comparator.GREATER_THAN_OR_EQUAL_NUMBER and user_value_float >= comparison_value_float):
                return True, error
        # IS ONE OF (hashed)
        elif comparator == Comparator.IS_ONE_OF_HASHED:
            user_value = self._get_user_attribute_value_as_text(comparison_attribute, user_value, condition, key)
            if sha256(encode_utf8(user_value), salt, context_salt) in comparison_value:
                return True, error
        # IS NOT ONE OF (hashed)
        elif comparator == Comparator.IS_NOT_ONE_OF_HASHED:
            user_value = self._get_user_attribute_value_as_text(comparison_attribute, user_value, condition, key)
            if sha256(encode_utf8(user_value), salt, context_salt) not in comparison_value:
                return True, error
        # BEFORE, AFTER (UTC datetime)
        elif Comparator.BEFORE_DATETIME <= comparator <= Comparator.AFTER_DATETIME:
            try:
                user_value_float = self._get_user_attribute_value_as_seconds_since_epoch(user_value)
            except ValueError:
                validation_error = "'%s' is not a valid Unix timestamp (number of seconds elapsed since Unix epoch)" % \
                                   str(user_value)
                error = self._handle_invalid_user_attribute(comparison_attribute, comparator, comparison_value, key,
                                                            validation_error)
                return False, error

            comparison_value_float = float(comparison_value)

            if (comparator == Comparator.BEFORE_DATETIME and user_value_float < comparison_value_float) \
                    or (comparator == Comparator.AFTER_DATETIME and user_value_float > comparison_value_float):
                return True, error
        # EQUALS (hashed)
        elif comparator == Comparator.EQUALS_HASHED:
            user_value = self._get_user_attribute_value_as_text(comparison_attribute, user_value, condition, key)
            if sha256(encode_utf8(user_value), salt, context_salt) == comparison_value:
                return True, error
        # NOT EQUALS (hashed)
        elif comparator == Comparator.NOT_EQUALS_HASHED:
            user_value = self._get_user_attribute_value_as_text(comparison_attribute, user_value, condition, key)
            if sha256(encode_utf8(user_value), salt, context_salt) != comparison_value:
                return True, error
        # STARTS WITH ANY OF, NOT STARTS WITH ANY OF, ENDS WITH ANY OF, NOT ENDS WITH ANY OF (hashed)
        elif Comparator.STARTS_WITH_ANY_OF_HASHED <= comparator <= Comparator.NOT_ENDS_WITH_ANY_OF_HASHED:
            for comparison in comparison_value:
                underscore_index = comparison.index('_')
                length = int(comparison[:underscore_index])
                user_value = self._get_user_attribute_value_as_text(comparison_attribute, user_value, condition, key)
                user_value_utf8 = encode_utf8(user_value)

                if len(user_value_utf8) >= length:
                    comparison_string = comparison[underscore_index + 1:]
                    if (comparator == Comparator.STARTS_WITH_ANY_OF_HASHED
                        and sha256(user_value_utf8[:length], salt, context_salt) == comparison_string) or \
                            (comparator == Comparator.ENDS_WITH_ANY_OF_HASHED
                             and sha256(user_value_utf8[-length:], salt, context_salt) == comparison_string):
                        return True, error
                    elif (comparator == Comparator.NOT_STARTS_WITH_ANY_OF_HASHED
                          and sha256(user_value_utf8[:length], salt, context_salt) == comparison_string) or \
                            (comparator == Comparator.NOT_ENDS_WITH_ANY_OF_HASHED
                             and sha256(user_value_utf8[-length:], salt, context_salt) == comparison_string):
                        return False, None

            # If no matches were found for the NOT_* conditions, then return True
            if comparator in [Comparator.NOT_STARTS_WITH_ANY_OF_HASHED, Comparator.NOT_ENDS_WITH_ANY_OF_HASHED]:
                return True, error
        # ARRAY CONTAINS ANY OF, ARRAY NOT CONTAINS ANY OF (hashed)
        elif Comparator.ARRAY_CONTAINS_ANY_OF_HASHED <= comparator <= Comparator.ARRAY_NOT_CONTAINS_ANY_OF_HASHED:
            try:
                user_value_list = self._get_user_attribute_value_as_string_list(user_value)

                if sys.version_info[0] == 2:
                    user_value_list = unicode_to_utf8(user_value_list)  # On Python 2.7, convert unicode to utf-8
            except ValueError:
                validation_error = "'%s' is not a valid string array" % str(user_value)
                error = self._handle_invalid_user_attribute(comparison_attribute, comparator, comparison_value, key,
                                                            validation_error)
                return False, error

            hashed_user_values = [sha256(encode_utf8(x), salt, context_salt) for x in user_value_list]
            if comparator == Comparator.ARRAY_CONTAINS_ANY_OF_HASHED:
                for comparison in comparison_value:
                    if comparison in hashed_user_values:
                        return True, error
            else:
                for comparison in comparison_value:
                    if comparison in hashed_user_values:
                        return False, None
                return True, error
        # EQUALS
        elif comparator == Comparator.EQUALS:
            user_value = self._get_user_attribute_value_as_text(comparison_attribute, user_value, condition, key)
            if user_value == comparison_value:
                return True, error
        # NOT EQUALS
        elif comparator == Comparator.NOT_EQUALS:
            user_value = self._get_user_attribute_value_as_text(comparison_attribute, user_value, condition, key)
            if user_value != comparison_value:
                return True, error
        # STARTS WITH ANY OF, NOT STARTS WITH ANY OF, ENDS WITH ANY OF, NOT ENDS WITH ANY OF
        elif Comparator.STARTS_WITH_ANY_OF <= comparator <= Comparator.NOT_ENDS_WITH_ANY_OF:
            user_value = self._get_user_attribute_value_as_text(comparison_attribute, user_value, condition, key)
            for comparison in comparison_value:
                if (comparator == Comparator.STARTS_WITH_ANY_OF and user_value.startswith(comparison)) or \
                        (comparator == Comparator.ENDS_WITH_ANY_OF and user_value.endswith(comparison)):
                    return True, error
                elif (comparator == Comparator.NOT_STARTS_WITH_ANY_OF and user_value.startswith(comparison)) \
                        or (comparator == Comparator.NOT_ENDS_WITH_ANY_OF and user_value.endswith(comparison)):
                    return False, None

            # If no matches were found for the NOT_* conditions, then return True
            if comparator in [Comparator.NOT_STARTS_WITH_ANY_OF, Comparator.NOT_ENDS_WITH_ANY_OF]:
                return True, error
        # ARRAY CONTAINS ANY OF, ARRAY NOT CONTAINS ANY OF
        elif Comparator.ARRAY_CONTAINS_ANY_OF <= comparator <= Comparator.ARRAY_NOT_CONTAINS_ANY_OF:
            try:
                user_value_list = self._get_user_attribute_value_as_string_list(user_value)

                if sys.version_info[0] == 2:
                    user_value_list = unicode_to_utf8(user_value_list)  # On Python 2.7, convert unicode to utf-8
            except ValueError:
                validation_error = "'%s' is not a valid string array" % str(user_value)
                error = self._handle_invalid_user_attribute(comparison_attribute, comparator, comparison_value, key,
                                                            validation_error)
                return False, error

            if comparator == Comparator.ARRAY_CONTAINS_ANY_OF:
                for comparison in comparison_value:
                    if comparison in user_value_list:
                        return True, error
            else:
                for comparison in comparison_value:
                    if comparison in user_value_list:
                        return False, None
                return True, error
        else:
            raise ValueError('Comparison operator is missing or invalid.')

        return False, error



import requests
import sys
from enum import IntEnum
from platform import python_version
from requests import HTTPError
from requests import Timeout

from .config import fixup_config_salt_and_segments
from .configentry import ConfigEntry
from .config import CONFIG_FILE_NAME, PREFERENCES, BASE_URL, REDIRECT
from .datagovernance import DataGovernance
from .logger import Logger
from .utils import get_utc_now_seconds_since_epoch, unicode_to_utf8
from .version import CONFIGCATCLIENT_VERSION

if sys.version_info < (2, 7, 9):
    requests.packages.urllib3.disable_warnings()

BASE_URL_GLOBAL = 'https://cdn-global.configcat.com'
BASE_URL_EU_ONLY = 'https://cdn-eu.configcat.com'
BASE_PATH = 'configuration-files/'
BASE_EXTENSION = '/' + CONFIG_FILE_NAME + '.json'


class RedirectMode(IntEnum):
    NoRedirect = 0,
    ShouldRedirect = 1,
    ForceRedirect = 2


class Status(IntEnum):
    Fetched = 0,
    NotModified = 1,
    Failure = 2


class FetchResponse(object):
    def __init__(self, status, entry, error=None, is_transient_error=False):
        self._status = status
        self.entry = entry
        self.error = error
        self.is_transient_error = is_transient_error

    def is_fetched(self):
        """Gets whether a new configuration value was fetched or not.
        :return: True if a new configuration value was fetched, otherwise False.
        """
        return self._status == Status.Fetched

    def is_not_modified(self):
        """Gets whether the fetch resulted a '304 Not Modified' or not.
        :return: True if the fetch resulted a '304 Not Modified' code, otherwise False.
        """
        return self._status == Status.NotModified

    def is_failed(self):
        """Gets whether the fetch failed or not.
        :return: True if the fetch failed, otherwise False.
        """
        return self._status == Status.Failure

    @classmethod
    def success(cls, entry):
        return FetchResponse(Status.Fetched, entry)

    @classmethod
    def not_modified(cls):
        return FetchResponse(Status.NotModified, ConfigEntry.empty)

    @classmethod
    def failure(cls, error, is_transient_error):
        return FetchResponse(Status.Failure, ConfigEntry.empty, error, is_transient_error)


class ConfigFetcher(object):
    def __init__(self, sdk_key, log, mode, base_url=None, proxies=None, proxy_auth=None,
                 connect_timeout=10, read_timeout=30, data_governance=DataGovernance.Global):
        self._sdk_key = sdk_key
        self.log = log
        self._proxies = proxies
        self._proxy_auth = proxy_auth
        self._connect_timeout = connect_timeout
        self._read_timeout = read_timeout
        self._headers = {'User-Agent': 'ConfigCat-Python/' + mode + '-' + CONFIGCATCLIENT_VERSION,
                         'X-ConfigCat-UserAgent': 'ConfigCat-Python/' + mode + '-' + CONFIGCATCLIENT_VERSION,
                         'X-ConfigCat-PythonVersion': python_version(),
                         'Content-Type': "application/json"}
        if base_url is not None:
            self._base_url_overridden = True
            self._base_url = base_url.rstrip('/')
        else:
            self._base_url_overridden = False
            if data_governance == DataGovernance.EuOnly:
                self._base_url = BASE_URL_EU_ONLY
            else:
                self._base_url = BASE_URL_GLOBAL

    def get_connect_timeout(self):
        return self._connect_timeout

    def get_read_timeout(self):
        return self._read_timeout

    def get_configuration(self, etag='', retries=0):
        """
        :return: Returns the FetchResponse object contains configuration entry
        """
        fetch_response = self._fetch(etag)

        # If there wasn't a config change, we return the response.
        if not fetch_response.is_fetched():
            return fetch_response

        preferences = fetch_response.entry.config.get(PREFERENCES, None)
        if preferences is None:
            return fetch_response

        base_url = preferences.get(BASE_URL)

        # If the base_url is the same as the last called one, just return the response.
        if base_url is None or self._base_url == base_url:
            return fetch_response

        redirect = preferences.get(REDIRECT)
        # If the base_url is overridden, and the redirect parameter is not 2 (force),
        # the SDK should not redirect the calls, and it just has to return the response.
        if self._base_url_overridden and redirect != int(RedirectMode.ForceRedirect):
            return fetch_response

        # The next call should use the base_url provided in the config json
        self._base_url = base_url

        # If the redirect property == 0 (redirect not needed), return the response
        if redirect == int(RedirectMode.NoRedirect):
            # Return the response
            return fetch_response

        # Try to download again with the new url

        if redirect == int(RedirectMode.ShouldRedirect):
            self.log.warning('The `dataGovernance` parameter specified at the client initialization is not in sync '
                             'with the preferences on the ConfigCat Dashboard. '
                             'Read more: https://configcat.com/docs/advanced/data-governance/',
                             event_id=3002)

        # To prevent loops we check if we retried at least 3 times with the new base_url
        if retries >= 2:
            self.log.error('Redirection loop encountered while trying to fetch config JSON. '
                           'Please contact us at https://configcat.com/support/', event_id=1104)
            return fetch_response

        # Retry the config download with the new base_url
        return self.get_configuration(etag, retries + 1)

    def _fetch(self, etag):  # noqa: C901
        uri = self._base_url + '/' + BASE_PATH + self._sdk_key + BASE_EXTENSION
        headers = self._headers
        if etag:
            headers['If-None-Match'] = etag
        else:
            headers['If-None-Match'] = None

        try:
            response = requests.get(uri, headers=headers, timeout=(self._connect_timeout, self._read_timeout),
                                    proxies=self._proxies, auth=self._proxy_auth)
            response.raise_for_status()

            if response.status_code in [200, 201, 202, 203, 204]:
                response_etag = response.headers.get('ETag')
                if response_etag is None:
                    response_etag = ''
                config = response.json()
                fixup_config_salt_and_segments(config)
                if sys.version_info[0] == 2:
                    config = unicode_to_utf8(config)  # On Python 2.7, convert unicode to utf-8
                    config_json_string = response.text.encode('utf-8')
                else:
                    config_json_string = response.text

                return FetchResponse.success(
                    ConfigEntry(config, response_etag, config_json_string, get_utc_now_seconds_since_epoch()))
            elif response.status_code == 304:
                return FetchResponse.not_modified()
            elif response.status_code in [404, 403]:
                error = 'Your SDK Key seems to be wrong. ' \
                        'You can find the valid SDK Key at https://app.configcat.com/sdkkey. ' \
                        'Received unexpected response: %s'
                error_args = (str(response), )
                self.log.error(error, *error_args, event_id=1100)
                return FetchResponse.failure(Logger.format(error, error_args), False)
            else:
                raise (requests.HTTPError(response))
        except HTTPError as e:
            error = 'Unexpected HTTP response was received while trying to fetch config JSON: %s'
            error_args = (str(e.response), )
            self.log.error(error, *error_args, event_id=1101)
            return FetchResponse.failure(Logger.format(error, error_args), True)
        except Timeout:
            error = 'Request timed out while trying to fetch config JSON. Timeout values: [connect: %gs, read: %gs]'
            error_args = (self.get_connect_timeout(), self.get_read_timeout())
            self.log.error(error, *error_args, event_id=1102)
            return FetchResponse.failure(Logger.format(error, error_args), True)
        except Exception as e:
            error = 'Unexpected error occurred while trying to fetch config JSON. It is most likely due to a local network ' \
                    'issue. Please make sure your application can reach the ConfigCat CDN servers (or your proxy server) ' \
                    'over HTTP.'
            self.log.exception(error, event_id=1103)
            return FetchResponse.failure(Logger.format(error, (), e), True)



from .configcatclient import ConfigCatClient
from .interfaces import ConfigCatClientException  # noqa: F401
from .datagovernance import DataGovernance  # noqa: F401
from .configcatoptions import ConfigCatOptions  # noqa: F401
from .pollingmode import PollingMode  # noqa: F401


def get(sdk_key, options=None):
    """
    Creates a new or gets an already existing `ConfigCatClient` for the given `sdk_key`.

    :param sdk_key: ConfigCat SDK Key to access your configuration.
    :param options: Configuration `ConfigCatOptions` for `ConfigCatClient`.
    :return: the `ConfigCatClient` instance.
    """
    return ConfigCatClient.get(sdk_key=sdk_key, options=options)


def close_all():
    """
    Closes all ConfigCatClient instances.
    """
    ConfigCatClient.close_all()



import sys
import inspect
from qualname import qualname
from datetime import datetime

try:
    from datetime import timezone
except ImportError:
    import pytz as timezone  # On Python 2.7, datetime.timezone is not available. We use pytz instead.

epoch_time = datetime(1970, 1, 1, tzinfo=timezone.utc)
distant_future = sys.float_info.max
distant_past = 0


def get_class_from_method(method):
    method_class = sys.modules.get(method.__module__)
    if method_class is None:
        return None
    for name in qualname(method).split('.')[:-1]:
        method_class = getattr(method_class, name)
    if not inspect.isclass(method_class):
        return None
    return method_class


def get_class_from_stack_frame(frame):
    args, _, _, value_dict = inspect.getargvalues(frame)
    # we check the first parameter for the frame function is
    # named 'self' or 'cls'
    if len(args):
        if args[0] == 'self':
            # in that case, 'self' will be referenced in value_dict
            instance = value_dict.get(args[0], None)
            if instance:
                # return its class
                return getattr(instance, '__class__', None)
        if args[0] == 'cls':
            # return the class
            return value_dict.get(args[0], None)

    # return None otherwise
    return None


def method_is_called_from(method, level=1):
    """
    Checks if the current method is being called from a certain method.
    """
    stack_info = inspect.stack()[level + 1]
    frame = stack_info[0]
    calling_method_name = frame.f_code.co_name
    expected_method_name = method.__name__
    if calling_method_name != expected_method_name:
        return False

    calling_class = get_class_from_stack_frame(frame)
    expected_class = get_class_from_method(method)
    if calling_class == expected_class:
        return True
    return False


def get_utc_now():
    return datetime.now(timezone.utc)


def get_seconds_since_epoch(date_time):
    # if there is no timezone info, assume UTC
    if date_time.tzinfo is None:
        date_time = date_time.replace(tzinfo=timezone.utc)

    return (date_time - epoch_time).total_seconds()


def get_date_time(seconds_since_epoch):
    return datetime.fromtimestamp(seconds_since_epoch, timezone.utc)


def get_utc_now_seconds_since_epoch():
    return get_seconds_since_epoch(get_utc_now())


def is_string_list(value):
    # Check if the value is a list
    if not isinstance(value, list):
        return False

    # Check if all items in the list are strings
    for item in value:
        # Handle unicode strings on Python 2.7
        if sys.version_info[0] == 2:
            if not isinstance(item, (str, unicode)):  # noqa: F821
                return False
        else:
            if not isinstance(item, str):
                return False

    return True


def unicode_to_utf8(data):
    """
    Convert unicode data in a collection to UTF-8 data. Used for supporting unicode config json strings on Python 2.7.
    Once Python 2.7 is no longer supported, this function can be removed.
    """
    if isinstance(data, dict):
        return {unicode_to_utf8(key): unicode_to_utf8(value) for key, value in data.iteritems()}
    elif isinstance(data, list):
        return [unicode_to_utf8(element) for element in data]
    elif isinstance(data, unicode):  # noqa: F821 (ignore warning: unicode is undefined in Python 3)
        return data.encode('utf-8')
    else:
        return data


def encode_utf8(value):
    """
    Get the UTF-8 encoded value of a string. Used for supporting unicode config json strings on Python 2.7.
    If the value is already UTF-8 encoded, it is returned as is.
    Once Python 2.7 is no longer supported, this function can be removed.
    The use of this function can be replaced with encode() method of the string: value.encode('utf-8')
    """
    try:
        return value.encode('utf-8')
    except UnicodeDecodeError:
        return value
